/*
 * An engine to intercept ajax requests
 * 
 * @class
 * @name RequestInterceptingEngine
 * @param {ZHPromise} ZHPromise - An instance of ZHPromise. 
 */
var RequestInterceptingEngine = function RequestInterceptingEngine(ZHPromise) {
	var interceptors = [];
	var REQUEST_ALL = "ALL";

	/*
	 * Returns all registered interceptors
	 * @function RequestInterceptingEngine.getInterceptors
	 * @returns {array} the registered interceptors
	 */
	this.getInterceptors = function getInterceptors() {
		return interceptors;
	};

	/*
	 * Registers a new interceptor
	 * @function RequestInterceptingEngine.registerInterceptor
	 * @param {object} interceptorConfiguration the configuration of the interceptor to be registered
	 * @returns {number} Index of the registered interceptor. Useful for deregistering later.
	 */
	this.registerInterceptor = function registerInterceptor(interceptorConfiguration) {
		var validation = validateInterceptorConfiguration(interceptorConfiguration);

		if (!validation.isValid) {
			throw new TypeError(validation.errorMessage);
		}

		return interceptors.push(interceptorConfiguration) - 1;
	};

	function validateInterceptorConfiguration(interceptorConfiguration) {
		var validator = new Validator();
		validator.initialMessage = "Failed to register interceptor. Parameter errors: ";

		var constraints = getInterceptorConfigurationConstraints(interceptorConfiguration);
		itemToValidate = [{ interceptorConfiguration: interceptorConfiguration }, interceptorConfiguration];
		validator.validate(itemToValidate, constraints);

		return validator;
	}

	function getInterceptorConfigurationConstraints() {
		var acceptedMethods = [REQUEST_ALL, RequestType.GET, RequestType.PUT, RequestType.POST, RequestType.DELETE, RequestType.PATCH];
		var requiredConstraint = {
			interceptorConfiguration: {
				presence: true,
				object: true
			}
		};

		var propertiesConstraint = {
			interceptor: {
				presence: true,
				function: true
			},
			method: {
				presence: true,
				stringInList: {
					stringList: acceptedMethods
				}
			}
		};

		return [requiredConstraint, propertiesConstraint];
	}

	/*
	 * Deregisters an interceptor form the given index.
	 * @function RequestInterceptingEngine.deregisterInterceptor
	 * @param {number} interceptorIndex the index of the interceptor to be deregistered.
	 */
	this.deregisterInterceptor = function deregisterInterceptor(interceptorIndex) {
		if (~interceptors.indexOf(interceptorIndex)) {
			interceptors.splice(interceptorIndex, 1);
		} else {
			ZhLogger.error("Unable to unregister interceptor. No interceptor found for the provided interceptorIndex (" + interceptorIndex + ")");
		}
	};

	/*
	 * Returns a promise from all the registered interceptors. If any of them rejects, the promise will be rejected.
	 * @function RequestInterceptingEngine.intercept
	 * @param {object|string} requestConfig the request configuration to be passed to thein terceptors
	 * @returns The promise from the interceptors result
	 */
	this.intercept = function intercept(requestConfig) {
		var interceptorsPromise;
		var requestMethodInterceptors = filterInterceptorsByMethod(interceptors, requestConfig.method);

		if (requestMethodInterceptors.length) {
			var interceptorsResult = interceptors.map(function (interceptorConfig) {
				var interceptor = interceptorConfig.interceptor;
				return interceptorConfig.scope ? interceptor.call(interceptorConfig.scope, requestConfig) : interceptor(requestConfig);
			});
			interceptorsPromise = ZHPromise.all(interceptorsResult);
		} else {
			//Use a thenableObject to make the result sync when there is no interceptor
			interceptorsPromise = thenableObject();
		}

		return interceptorsPromise;
	};

	function filterInterceptorsByMethod(interceptors, method) {
		return interceptors.filter(function (interceptor) {
			return interceptor.method === REQUEST_ALL || interceptor.method === method;
		});
	}

	function thenableObject() {
		var thenable = {};
		thenable.then = function (callback) {
			if (callback) {
				callback();
			}
			return thenable;
		};
		thenable.catch = function () {
			return thenable;
		};

		return thenable;
	}
};